import copy

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib.backends.backend_pdf import PdfPages
from scipy import linalg

import A_FMM
import A_FMM.sub_sm as sub
from A_FMM.creator import Creator
from A_FMM.scattering import S_matrix


class Layer:
    """Class for the definition of a single layer"""

    def __init__(self, Nx: int, Ny: int, creator: Creator, Nyx: float = 1.0):
        """Creator

        Args:
            Nx (int): truncation order in x direction
            Ny (int): truncation order in y direction
            Nyx (float): ratio between the cell's dimension in y and x (ay/ax)
        """
        self.Nx = Nx
        self.Ny = Ny
        self.NPW = (2 * Nx + 1) * (2 * Ny + 1)
        self.G = sub.createG(self.Nx, self.Ny)
        self.G_inv = {v: k for k, v in self.G.items()}
        self.D = len(self.G)
        self.creator = copy.deepcopy(creator)
        self.Nyx = Nyx

        self.FOUP = self.__create_eps()
        self.INV = linalg.inv(self.FOUP)

        self.EPS1 = sub.fou_xy(
            self.Nx,
            self.Ny,
            self.G,
            self.creator.x_list,
            self.creator.y_list,
            self.creator.eps_lists,
        )
        self.EPS2 = sub.fou_yx(
            self.Nx,
            self.Ny,
            self.G,
            self.creator.x_list,
            self.creator.y_list,
            self.creator.eps_lists,
        )

        self.TX = False
        self.TY = False

    def __create_eps(self):
        nx = 2 * self.Nx
        ny = 2 * self.Ny
        mx = 4 * self.Nx + 1
        my = 4 * self.Ny + 1
        fourier_transform = np.zeros((mx, my), dtype=complex)
        x_list = self.creator.x_list
        y_list = self.creator.y_list
        eps_lists = self.creator.eps_lists
        G = self.G
        for i in range(mx):
            for j in range(my):
                fourier_transform[i, j] = sub.fou(
                    (i + nx) % mx - nx, (j + ny) % my - ny, x_list, y_list, eps_lists
                )
        D = len(G)
        F = np.zeros((D, D), complex)
        for i, (gx1, gy1) in G.items():
            for j, (gx2, gy2) in G.items():
                F[i, j] = fourier_transform[gx1 - gx2, gy1 - gy2]
        return F

    def inspect(self, st=""):
        """Function for inspectig the attributes of a layer object

        Args:
            st (string): string to print before the inspection for identification
        """
        att = sub.get_user_attributes(self)
        print(st)
        print(22 * "_")
        print("| INT argument")
        for i in att:
            if type(i[1]) is int:
                print("|%10s%10s" % (i[0], str(i[1])))
        print("| Float argument")
        for i in att:
            if type(i[1]) is float:
                print("|%10s%10s" % (i[0], str(i[1])))
        for i in att:
            if type(i[1]) is np.float64:
                print("|%10s%10s" % (i[0], str(i[1])))
        print("| BOOL argument")
        for i in att:
            if type(i[1]) is bool:
                print("|%10s%10s" % (i[0], str(i[1])))
        print("| Array argument")
        for i in att:
            if type(i[1]) is np.ndarray:
                print("|%10s%10s" % (i[0], str(np.shape(i[1]))))
        print("")

    def eps_plot(self, pdf=None, N=200, s=1):
        """Function for plotting the dielectric consstat rebuit from plane wave expansion

        Args:
            pdf (string or PdfPages): file for printing the the epsilon
                if a PdfPages object, the page is appended to the pdf
                if string, a pdf with that name is created
            N (int): number of points
            s (float): number of cell replicas to display (default 1)
        """
        [X, Y] = np.meshgrid(
            np.linspace(-s * 0.5, s * 0.5, s * N),
            np.linspace(-s * 0.5, s * 0.5, int(s * N * self.Nyx)),
        )
        EPS = np.zeros((N, N), complex)
        #        for i in range(self.D):
        #            EPS+=sub.fou(self.G[i][0],self.G[i][1],self.creator.x_list,self.creator.y_list,self.creator.eps_lists)*np.exp(-(0+2j)*np.pi*(self.G[i][0]*X+self.G[i][1]*Y))
        EPS = sum(
            [
                sub.fou(
                    self.G[i][0],
                    self.G[i][1],
                    self.creator.x_list,
                    self.creator.y_list,
                    self.creator.eps_lists,
                )
                * np.exp((0 + 2j) * np.pi * (self.G[i][0] * X + self.G[i][1] * Y))
                for i in range(self.D)
            ]
        )
        plt.figure()
        # plt.imshow(np.real(EPS),aspect='auto',extent=[-s*0.5,s*0.5,-self.Nyx*s*0.5,self.Nyx*s*0.5])
        plt.imshow(
            np.real(EPS),
            extent=[-s * 0.5, s * 0.5, -self.Nyx * s * 0.5, self.Nyx * s * 0.5],
            origin="lower",
        )
        plt.colorbar()
        if pdf == None:
            plt.show()
        elif isinstance(pdf, PdfPages):
            pdf.savefig()
        else:
            a = PdfPages(pdf + ".pdf")
            a.savefig()
            a.close()
        plt.close()

    def transform(self, ex: float = 0, ey: float = 0, complex_transform: bool = False):
        """Function for adding the real coordinate transfomr to the layer

        Note: for no mapping, set the width to 0

        Args:
            ex (float): relative width of the unmapped region in x direction. Default is 0 (no mapping)
            ey (float): relative width of the unmapped region in y direction. Default is 0 (no mapping)
            complex_transform (bool): False for real transform (default), True for complex one.
        """
        if complex_transform:
            transform_function = sub.fou_complex_t
        else:
            transform_function = sub.fou_t

        if ex != 0.0:
            self.TX = True
            self.ex = ex
            self.FX = np.zeros((self.D, self.D), complex)
            nx = 2 * self.Nx
            mx = 4 * self.Nx + 1
            F = [transform_function((i + nx) % mx - nx, ex) for i in range(mx)]
            for i, (gx1, gy1) in self.G.items():
                for j, (gx2, gy2) in self.G.items():
                    if gy1 != gy2:
                        continue
                    self.FX[i, j] = F[gx1 - gx2]
        else:
            self.FX = None

        if ey != 0.0:
            self.TY = True
            self.ey = ey
            self.FY = np.zeros((self.D, self.D), complex)
            ny = 2 * self.Ny
            my = 4 * self.Ny + 1
            F = [transform_function((i + ny) % my - ny, ey) for i in range(my)]
            for i, (gx1, gy1) in self.G.items():
                for j, (gx2, gy2) in self.G.items():
                    if gx1 != gx2:
                        continue
                    self.FY[i, j] = F[gy1 - gy2]
        else:
            self.FY = None

        return self.FX, self.FY

    def add_transform_matrix(
        self,
        ex: float = 0.0,
        FX: np.ndarray = None,
        ey: float = 0.0,
        FY: np.ndarray = None,
    ):
        """Function for adding matrix of a coordinate transform

        Args:
            ex (float): relative width of the unmapped region in x direction. Default is 0. This is only for keeping track of the value, as it has no effect on the transformation.
            FX (ndarray): FX matrix of the coordinate trasnformation
            ey (float): relative width of the unmapped region in y direction. Default is 0. This is only for keeping track of the value, as it has no effect on the transformation.
            FY (ndarray): FY matrix of the coordinate trasnformation
        """
        if ex != 0:
            self.TX = True
            self.ex = ex
            self.FX = FX
        else:
            self.FX = None
        if ey != 0:
            self.TY = True
            self.ey = ey
            self.FY = FY
        else:
            self.FY = None

    def mode(self, k0: float, kx: float = 0.0, ky: float = 0.0):
        """Calculates the eighenmode of the layer

        Args:
            k0 (float): Vacuum wavevector
            kx (float): Wavevector in the x direction
            ky (float): Wavevector in the y direction
        """
        self.k0 = k0
        self.kx = kx
        self.ky = ky
        (k1, k2) = sub.createK(self.G, k0, kx=kx, ky=ky, Nyx=self.Nyx)
        if self.TX:
            k1 = np.dot(self.FX, k1)
        if self.TY:
            k2 = np.dot(self.FY, k2)
        self.GH, self.M = sub.create_2order_new(
            self.D, k1, k2, self.INV, self.EPS1, self.EPS2
        )
        [self.W, self.V] = linalg.eig(self.M)
        self.gamma = np.sqrt(self.W) * np.sign(np.angle(self.W) + 0.5 * np.pi)
        if np.any(np.real(self.gamma) + np.imag(self.gamma) <= 0.0):
            print("Warining: wrong complex root")
        if np.any(np.abs(self.gamma) <= 0.0):
            print("Warining: gamma=0")
        self.VH = np.dot(self.GH, self.V / self.gamma)

    def clear(self):
        """Removes data created in mode method"""
        self.VH = None
        self.M = None
        self.GH = None
        self.W = None
        self.V = None
        self.gamma = None

    def get_index(self, ordered: bool = True) -> np.ndarray:
        """Returns the effective idexes of the modes

        Args:
            ordered (bool): if True (default) the modes are ordered by decreasing effective index
        """
        if ordered:
            Neff = self.gamma[np.argsort(self.W)[::-1]]
        else:
            Neff = self.gamma
        return Neff

    def mat_plot(self, name: str):
        """Plot the absolute values of the fourier trasnsform matrices

        Args:
            name (str): name of the pdf file for plotting
            N (int): number of points for plotting the epsilon
            s (float): number pf relicas of the cell to plot. Default is 1.
        """
        with PdfPages(name + ".pdf") as save:
            for attr in ["FOUP", "EPS1", "EPS2", "INV", "FX", "FY"]:
                try:
                    to_plot = getattr(self, attr)
                    plt.figure()
                    plt.title(attr)
                    plt.imshow(np.abs(to_plot), aspect="auto", interpolation="nearest")
                    plt.colorbar()
                    save.savefig()
                    plt.close()
                except AttributeError:
                    pass

    def plot_Ham(self, pdf: PdfPages) -> None:
        """Plot the matrix of the eigenvalue problem

        Args:
            pdf (PdfPages): pdf object to be used to plot.

        Returns:
            None.

        """
        plt.figure()
        plt.title("k0:%5.3f kx:%5.3f ky:%5.3f" % (self.k0, self.kx, self.ky))
        plt.imshow(np.abs(np.abs(self.M)), aspect="auto", interpolation="nearest")
        plt.colorbar()
        pdf.savefig()
        plt.close()

    def _process_xy(self, x: np.ndarray, y: np.ndarray) -> tuple:
        """Transform the x and y coordinates between the real and computational space

        Args:
            x (ndarray): array of x coordinates in the real space
            y (ndarray): array of y coordinates in the real space

        Returns: tuple of numpy.ndarray containing:
            - ndarray: array of x coordinates in the computational space
            - ndarray: array of y coordinates in the computational space

        """
        if self.TX:
            x = sub.t_inv(x, self.ex)
        if self.TY:
            y = sub.t_inv(y, self.ey)
        return x, y

    def calculate_epsilon(
        self, x: np.ndarray = 0.0, y: np.ndarray = 0.0, z: np.ndarray = 0.0
    ) -> dict[str, np.ndarray]:
        """Return epsilon given the coordinates

        The epsilon returned here is the one reconstructed from the Fourier transform.
        The epsilon is reconstructed on the meshgrid of x,y, and z.

        Args:
            x (array_like): x coordinates (1D array).
            y (array_like): y coordinates (1D array).
            z (array_like): z coordinates (1D array).

        Returns:
            ndarray : Epsilon value at coordinates. Shape of ndarray is the same as x,y, and z.

        Raises:
            ValueError:
                if x,y,z have different shapes
        """

        x, y = self._process_xy(x, y)
        eps = self.FOUP[:, self.D // 2]
        gx, gy = zip(*[g for i, g in self.G.items()])
        gx, gy = np.asarray(gx), np.asarray(gy)
        xp, yp, gxp = np.meshgrid(x, y, gx, indexing="ij")
        xp, yp, gyp = np.meshgrid(x, y, gy, indexing="ij")
        eps_p = np.dot(np.exp((0 + 2j) * np.pi * (gxp * xp + gyp * yp)), eps)
        shape = np.shape(eps_p)
        EPS, _ = np.meshgrid(eps_p, z, indexing="ij")
        EPS = EPS.reshape(*shape, -1)
        x, y, z = np.meshgrid(x, y, z, indexing="ij")
        eps = {
            "x": x,
            "y": y,
            "z": z,
            "eps": EPS,
        }
        return eps

    @staticmethod
    def _filter_componets(components: list = None) -> list:
        """
        Checks if the fileds components list contains only allowed ones
        """
        if components is None:
            return ["Ex", "Ey", "Hx", "Hy"]
        for comp in components:
            if comp not in ["Ex", "Ey", "Hx", "Hy"]:
                raise ValueError(
                    f"Field component f{comp} not available. Only Ex, Ey, Hx, or Hy are allowed"
                )
        return components

    @staticmethod
    def _check_array_shapes(u: np.ndarray, d: np.ndarray) -> None:
        """
        Chekcs that the modal amplitudea arrays and the coordinates arrays have consistent shapes
        """
        if np.shape(u) != np.shape(d):
            raise ValueError(
                f"Shape of u different from shape of d {np.shape(u)}!={np.shape(d)}"
            )

    def calculate_field_old(
        self,
        u: np.ndarray,
        d: np.ndarray = None,
        x: np.ndarray = 0,
        y: np.ndarray = 0,
        z: np.ndarray = 0,
        components: list = None,
    ) -> dict:
        """Return field given modal coefficient and coordinates

        Coordinates arrays must be 1D. Fields are returned on a meshgrid of the input coordinates.
        Older version. Slower, but may require less memory.

        Args:
            u (array_like): coefficient of forward propagating modes.
            d (array_like, optional): coefficient of backward propagating modes.
                Default to None: no backward propagation is assumed.
            x (array_like): x coordinates.
            y (array_like): y coordinates.
            z (array_like): z coordinates.
            components (list of str, optional): field components to calculate.
                Default to None: all components ('Ex', 'Ey', 'Hx', 'Hy') are calculated.

        Returns:
            dict of ndarray : Desired field components. Shape of ndarray is the same as x,y, and z.

        Raises:
            ValueError:
                if other component than 'Ex', 'Ey', 'Hx', or 'Hy' is requested.

        """
        components = self._filter_componets(components)
        d = np.zeros_like(u, dtype=complex) if d is None else d
        self._check_array_shapes(u, d)

        x, y = self._process_xy(x, y)
        x, y, z = np.meshgrid(x, y, z, indexing="ij")
        field = {
            "x": x,
            "y": y,
            "z": z,
        }
        field.update({comp: np.zeros_like(x, dtype=complex) for comp in components})
        for i, (uu, dd, n) in enumerate(zip(u, d, self.gamma)):
            if uu == 0.0 and dd == 0.0:
                continue
            field_tmp = {comp: np.zeros_like(x, dtype=complex) for comp in components}
            for j, (gx, gy) in self.G.items():
                [WEx, WEy] = np.split(self.V[:, i], 2)
                [WHx, WHy] = np.split(self.VH[:, i], 2)
                EXP = np.exp(
                    (0 + 2j) * np.pi * ((gx + self.kx) * x + (gy + self.ky) * y)
                )
                for comp in components:
                    sign = 1.0 if comp[0] == "E" else -1.0
                    coeff = uu * np.exp(
                        2.0j * np.pi * self.k0 * n * z
                    ) + sign * dd * np.exp(-2.0j * np.pi * self.k0 * n * z)
                    field_tmp[comp] = (
                        field_tmp[comp] + coeff * eval(f"W{comp}")[j] * EXP
                    )
            for comp in components:
                field[comp] = field[comp] + field_tmp[comp]
        return field

    def calculate_field(
        self,
        u: np.ndarray,
        d: np.ndarray = None,
        x: np.ndarray = 0,
        y: np.ndarray = 0,
        z: np.ndarray = 0,
        components: list = None,
    ) -> dict:
        """Return field given modal coefficient and coordinates

        Coordinates arrays must be 1D. Fields are returned on a meshgrid of the input coordinates.

        Args:
            u (array_like): coefficient of forward propagating modes.
            d (array_like, optional): coefficient of backward propagating modes.
                Default to None: no backward propagation is assumed.
            x (array_like): x coordinates.
            y (array_like): y coordinates.
            z (array_like): z coordinates.
            components (list of str, optional): field components to calculate.
                Default to None: all components ('Ex', 'Ey', 'Hx', 'Hy') are calculated.

        Returns:
            dict of ndarray : Desired field components. Shape of ndarray is the same as x,y, and z.

        Raises:
            ValueError:
                if other component than 'Ex', 'Ey', 'Hx', or 'Hy' is requested.

        """
        x, y = self._process_xy(x, y)
        components = self._filter_componets(components)
        d = np.zeros_like(u, dtype=complex) if d is None else d
        self._check_array_shapes(u, d)
        X, Y, Z = np.meshgrid(x, y, z, indexing="ij")
        field = {
            "x": X,
            "y": Y,
            "z": Z,
        }
        Gx = [gs[0] for (i, gs) in self.G.items()]
        Gy = [gs[1] for (i, gs) in self.G.items()]
        u, d = np.asarray(u), np.asarray(d)
        ind = [i for i, (uu, dd) in enumerate(zip(u, d)) if uu != 0.0 or dd != 0.0]
        u = u[ind]
        d = d[ind]
        WEx, WEy = np.split(self.V, 2, axis=0)
        WHx, WHy = np.split(self.VH, 2, axis=0)
        W = {
            "Ex": WEx[:, ind],
            "Ey": WEy[:, ind],
            "Hx": WHx[:, ind],
            "Hy": WHy[:, ind],
        }
        X, Y, Gx = np.meshgrid(x, y, Gx, indexing="ij")
        X, Y, Gy = np.meshgrid(x, y, Gy, indexing="ij")
        EXP = np.exp(2.0j * np.pi * ((Gx + self.kx) * X + (Gy + self.ky) * Y))
        u, Z = np.meshgrid(u, z, indexing="ij")
        d, Z = np.meshgrid(d, z, indexing="ij")
        n, Z = np.meshgrid(self.gamma[ind], z, indexing="ij")
        z_exp = 2.0j * np.pi * self.k0 * n * Z
        coeff_u = u * np.exp(z_exp)
        coeff_d = d * np.exp(-z_exp)
        for comp in components:
            coeff = coeff_u + coeff_d if "E" in comp else coeff_u - coeff_d
            EXPV = np.dot(EXP, W[comp])
            field[comp] = np.dot(EXPV, coeff)
        return field

    def get_modal_field(
        self, i: int, x: float = 0.0, y: float = 0.0, components: list = None
    ) -> dict:
        """Returns modal field profile

        Args:
            i (int): index of the mode.
            x (float or array_like): x coordinate for the field calculation
            y (float or array_like): y coordinate for the field calculation
            components (list of str, optional): field components to calculate.
                Default to None: all components ('Ex', 'Ey', 'Hx', 'Hy') are calculated.

        Returns:
            DataFrame: modal field
        """
        x, y, z = np.meshgrid(x, y, [0.0], indexing="ij")
        u = self.create_input({i: 1.0})
        data = {
            "x": np.squeeze(x),
            "y": np.squeeze(y),
        }
        field = self.calculate_field(x, y, z, u, components=components)
        for k, v in field.items():
            data[k] = np.squeeze(v)
        return data

    def get_P_norm(self):
        """Creates array of single mode Poynting vector components.

        It is stored in the P_norm attribute

        Returns:
            None.

        """
        [VEx, VEy] = np.split(self.V, 2)
        [VHx, VHy] = np.split(self.VH, 2)
        self.P_norm = np.sum(VEx * np.conj(VHy) - VEy * np.conj(VHx), 0).real

    def get_Poynting_single(self, i: int, u: np.ndarray, ordered: bool = True) -> float:
        """Return the Poyinting vector of a single mode given the modal expansion in the layer


        Args:
            i (int): Index of the mode.
            u (1darray): Array of modal coefficient.
            ordered (TYPE, optional): Regulates how mode are ordered. If True, they are ordered for decreasing effective index. If Flase, the order is whatever is returned by the diagonalization routine. Defaults to True.

        Returns:
            TYPE: DESCRIPTION.

        """
        if ordered:
            j = np.argsort(self.W)[-i - 1]
        else:
            j = i
        self.get_Poyinting_norm()
        return self.PP_norm[j, j].real * np.abs(u[j]) ** 2.0

    def get_Poyinting_norm(self):
        """Calculates the normalization matrix for the Poyinting vector calculations


        Returns:
            None.

        """
        [VEx, VEy] = np.split(self.V, 2)
        [VHx, VHy] = np.conj(np.split(self.VH, 2))
        # old version (working)
        # self.PP_norm=np.zeros((2*self.D,2*self.D),dtype=complex)
        # for i in range(self.D):
        #    VEX,VHY=np.meshgrid(VEx[i,:],VHy[i,:])
        #    VEY,VHX=np.meshgrid(VEy[i,:],VHx[i,:])
        #    P1=np.multiply(VEX,VHY)
        #    P2=-np.multiply(VEY,VHX)
        #    P=np.add(P1,P2)
        #    self.PP_norm=np.add(self.PP_norm,P)
        # print self.PP_norm
        # new version. should be equivalent bit faster
        P1 = np.dot(np.transpose(VEx), VHy)
        P2 = np.dot(np.transpose(VEy), VHx)
        self.PP_norm = np.add(P1, -P2)

    def get_Poynting(self, u: np.ndarray, d: np.ndarray = None):
        """Calculates total Poynting vector in the layer given arrays of modal expansion


        Args:
            u (1darray): Modal expansion of forward propagating modes.
            d (1darray, optional): Modal expansion of backward propagating modes. Defaults to None.

        Returns:
            TYPE: DESCRIPTION.

        """
        if d is None:
            d = np.zeros(2 * self.D, dtype=complex)
        # try:
        #    self.PP_norm
        # except AttributeError:
        #    self.get_Poyinting_norm()
        self.get_Poyinting_norm()
        Cn = np.add(u, d)
        Cnp = np.add(u, -d)
        [Cn, Cnp] = np.meshgrid(Cn, np.conj(Cnp))
        C = np.multiply(Cn, Cnp)
        PP = np.multiply(C, self.PP_norm)
        return np.sum(PP).real

    def T_interface(self, lay) -> np.ndarray:
        """Builds the Transfer matrix of the interface with another layer

        Args:
            lay (Layer): Layer toward which to calculate the scattering matrix.

        Returns:
            T (2darray): Interface scattering matrix.

        """
        T1 = np.dot(linalg.inv(lay.V), self.V)
        T2 = np.dot(linalg.inv(lay.VH), self.VH)
        T11 = 0.5 * (T1 + T2)
        T12 = 0.5 * (T1 - T2)
        T21 = 0.5 * (T1 - T2)
        T22 = 0.5 * (T1 + T2)
        T = np.vstack([np.hstack([T11, T12]), np.hstack([T21, T22])])
        return T

    def T_prop(self, d: float) -> np.ndarray:
        """Build the propagation Transfer matrix of the layer

        Args:
            d (float): Thickness of the layer.

        Returns:
            T (2darray): Propagation Transfer matrix.

        """
        I1 = np.diag(np.exp((0 + 1j) * self.k0 * self.gamma * d))
        I2 = np.diag(np.exp(-(0 + 1j) * self.k0 * self.gamma * d))
        I = np.zeros((2 * self.D, 2 * self.D), complex)
        T = np.vstack([np.hstack([I1, I]), np.hstack([I, I2])])
        return T

    # newer version, should be faster
    def interface(self, lay) -> S_matrix:
        """Builds the Scattering matrix of the interface with another layer

        Args:
            lay (Layer): Layer toward which to calculate the scattering matrix.

        Returns:
            S (S_matrix): Interface scattering matrix.

        """
        S = S_matrix(2 * self.D)
        T1 = np.dot(linalg.inv(lay.V), self.V)
        T2 = np.dot(linalg.inv(lay.VH), self.VH)
        T11 = 0.5 * (T1 + T2)
        T12 = 0.5 * (T1 - T2)
        # T21= 0.5*(T1 - T2)
        # T22= 0.5*(T1 + T2)
        # T=np.vstack([np.hstack([T11,T12]),np.hstack([T21,T22])])
        Tm = linalg.inv(T11)
        S.S11 = T11 - np.dot(np.dot(T12, Tm), T12)
        S.S12 = np.dot(T12, Tm)
        S.S21 = -np.dot(Tm, T12)
        S.S22 = Tm
        return S

    def get_input(
        self,
        func: callable,
        args: tuple = None,
        Nxp: int = 1024,
        Nyp: int = None,
        fileprint: str = None,
    ) -> np.ndarray:
        """Expands an arbitrary fieldd shape on the basis of the layer eigenmodes

        Args:
            func (function): Function describing the field.
                This function should be in the form (x,y,*args). It must be able to accept x and y as numpy array.
                It must return two values, expressing Ex and Ey
            args (tuple, optional): Eventual tuple of additional arguments for func. Defaults to None.
            Nxp (int, optional): Number of points to evaluate the function in the x direction. Defaults to 1024.
            Nyp (int, optional): Number of points to evaluate the function in the y direction. Defaults to None (1 if layer is 1D, Nxp if 2D).
            fileprint (str, optional): Filename on which to write the used function. Mainly for debug. Defaults to None.

        Returns:
            u (1darray): Array of the modal coefficient of the expansion.

        """
        args = () if args is None else args

        if Nyp == None:
            if self.Ny == 0:
                Nyp = 1
                y = np.array([0.0])
            else:
                Nyp = Nxp
                y = np.linspace(-0.5, 0.5, Nyp)
        else:
            y = np.linspace(-0.5, 0.5, Nyp)
        x = np.linspace(-0.5, 0.5, Nxp)
        if self.TX:
            ex = self.ex
            x = sub.t_dir(x, ex)
        if self.TY:
            ey = self.ey
            y = sub.t_dir(y, ey)

        y = y * self.Nyx

        [X, Y] = np.meshgrid(x, y, indexing="ij")
        [Fx, Fy] = func(X, Y, *args)

        Fx = np.fft.fftshift(Fx) / (Nxp * Nyp)
        Fy = np.fft.fftshift(Fy) / (Nxp * Nyp)

        FOUx = np.fft.fft2(Fx)
        FOUy = np.fft.fft2(Fy)

        Estar = np.zeros(2 * self.NPW, dtype=complex)
        for i in range(self.NPW):
            # print self.G[i][0], self.G[i][1],FOUx[self.G[i][0],self.G[i][1]]
            Estar[i] = FOUx[self.G[i][0], self.G[i][1]]
            Estar[i + self.NPW] = FOUy[self.G[i][0], self.G[i][1]]

        u = linalg.solve(self.V, Estar)
        return u

    def create_input(self, dic: dict) -> np.ndarray:
        """Creates the array of modal coefficient using a dictionary as input


        Args:
            dic (dict): Dictionary of exited modes {modal_index : modal_coeff}. Modes are ordered.

        Returns:
            u (1darray): Array of modal coefficient.

        """
        u = np.zeros((2 * self.NPW), complex)
        for i in dic:
            u[np.argsort(self.W)[-i - 1]] = dic[i]
        return u

    def get_Enorm(self):
        """Calculate field normalization


        Returns:
            None.

        """
        [VEx, VEy] = np.split(self.V, 2)
        self.ENx = np.dot(np.transpose(VEx), np.conj(VEx))
        self.ENy = np.dot(np.transpose(VEy), np.conj(VEy))

    def overlap(self, u: np.ndarray, up: np.ndarray = None):
        """EXPERIMENTAL: Calculates overlap between two fields given the modal expansion

        Args:
            u (1darray): Modal coefficient of first mode.
            up (1darray, optional): Modal coefficient of first mode. Defaults to None (up=u, namely normalization is returned).

        Returns:
            list: [tx, tx]: floats. Namely overlap in x and y polarization

        """
        if up is None:
            up = u
        try:
            self.ENx
        except AttributeError:
            self.get_Enorm()
        # print np.shape(u),np.shape(self.ENx)
        tx = np.dot(self.ENx, np.conj(up))
        ty = np.dot(self.ENy, np.conj(up))
        tx = np.dot(np.transpose(u), tx)
        ty = np.dot(np.transpose(u), ty)
        # print tx,ty
        return [tx, ty]

    def coupling(self, u: np.ndarray, up: np.ndarray) -> tuple:
        """EXPERIMENTAL: Calculates coupling between two modes given their modal exapnsion

        Args:
            u (TYPE): Modal coefficient of first mode.
            up (TYPE): Modal coefficient of second mode.

        Returns:
            list: [tx, tx]: floats. Coupling in x and y polarization.

        """
        self.get_Enorm()
        [tx1, ty1] = self.overlap(u)
        [tx2, ty2] = self.overlap(up)
        [txc, tyc] = self.overlap(u, up)
        return txc / np.sqrt(tx1 * tx2), tyc / np.sqrt(ty1 * ty2)


class Layer_ani_diag(Layer):
    """Class for the definition of a single layer anysitropic (diagonal) layer"""

    def __init__(
        self,
        Nx: int,
        Ny: int,
        creator_x: Creator,
        creator_y: Creator,
        creator_z: Creator,
        Nyx: float = 1.0,
    ):
        self.Nx = Nx
        self.Ny = Ny
        self.NPW = (2 * Nx + 1) * (2 * Ny + 1)
        self.G = sub.createG(self.Nx, self.Ny)
        self.D = len(self.G)
        self.creator = [
            copy.deepcopy(creator_x),
            copy.deepcopy(creator_y),
            copy.deepcopy(creator_z),
        ]
        self.Nyx = Nyx

        self.FOUP = sub.create_epsilon(
            self.G,
            self.creator[2].x_list,
            self.creator[2].y_list,
            self.creator[2].eps_lists,
        )
        self.INV = linalg.inv(self.FOUP)

        self.EPS1 = sub.fou_xy(
            self.Nx,
            self.Ny,
            self.G,
            self.creator[0].x_list,
            self.creator[0].y_list,
            self.creator[0].eps_lists,
        )
        self.EPS2 = sub.fou_yx(
            self.Nx,
            self.Ny,
            self.G,
            self.creator[1].x_list,
            self.creator[1].y_list,
            self.creator[1].eps_lists,
        )

        self.TX = False
        self.TY = False


class Layer_num(Layer):
    """Class for the definition of a single layer from a function defining the dielectric profile"""

    def __init__(
        self,
        Nx: int,
        Ny: int,
        func: callable,
        args: tuple = None,
        Nyx: float = 1.0,
        NX: int = 2048,
        NY: int = 2048,
    ):
        self.Nx = Nx
        self.Ny = Ny
        self.NPW = (2 * Nx + 1) * (2 * Ny + 1)
        self.G = sub.createG(self.Nx, self.Ny)
        self.D = len(self.G)
        self.Nyx = Nyx
        self.func = func
        self.args = args
        self.NX = NX
        self.NY = NY
        # print args
        args = () if args is None else args

        self.FOUP = sub.num_fou(func, args, self.G, NX, NY, self.Nyx)
        self.INV = linalg.inv(self.FOUP)

        # Still to be defined
        self.EPS1 = sub.num_fou_xy(
            self.func, self.args, self.Nx, self.Ny, self.G, NX, NY, self.Nyx
        )
        self.EPS2 = sub.num_fou_yx(
            self.func, self.args, self.Nx, self.Ny, self.G, NX, NY, self.Nyx
        )

        self.TX = False
        self.TY = False


class Layer_uniform(Layer):
    """Class for the definition of a single uniform layer"""

    def __init__(self, Nx: int, Ny: int, eps: float | complex, Nyx: float = 1.0):
        self.Nx = Nx
        self.Ny = Ny
        self.NPW = (2 * Nx + 1) * (2 * Ny + 1)
        self.G = sub.createG(self.Nx, self.Ny)
        self.D = len(self.G)
        self.Nyx = Nyx
        self.eps = eps

        self.FOUP = eps * np.identity(self.D, dtype="complex")
        self.INV = 1.0 / eps * np.identity(self.D, dtype="complex")

        # Still to be defined
        self.EPS1 = eps * np.identity(self.D, dtype="complex")
        self.EPS2 = eps * np.identity(self.D, dtype="complex")

        self.TX = False
        self.TY = False


class Layer_empty_st(Layer):
    """Class for the definition of an empy layer"""

    def __init__(self, Nx: int, Ny: int, creator: Creator, Nyx: float = 1.0):
        self.Nx = Nx
        self.Ny = Ny
        self.NPW = (2 * Nx + 1) * (2 * Ny + 1)
        self.G = sub.createG(self.Nx, self.Ny)
        self.D = len(self.G)
        self.creator = copy.deepcopy(creator)
        self.Nyx = Nyx

        self.TX = False
        self.TY = False

        self.FOUP = np.zeros((self.D, self.D), dtype=complex)
        # self.INV=np.zeros((self.D,self.D),dtype=complex)
        self.INV = linalg.inv(np.eye(self.D, dtype=complex))
        self.EPS1 = np.zeros((self.D, self.D), dtype=complex)
        self.EPS2 = np.zeros((self.D, self.D), dtype=complex)

    def fourier(self):
        """Calculates the fourier transform matrices need for the eigenvalue problem.


        Returns:
            2darray: FOUP matrix.
            2darray: INV matrix.
            2darray: EPS1 matrix.
            2darray: EPS2 matrix.

        """
        self.FOUP = sub.create_epsilon(
            self.G, self.creator.x_list, self.creator.y_list, self.creator.eps_lists
        ) * (1.0 + 0.0j)
        self.INV = linalg.inv(self.FOUP)
        self.EPS1 = sub.fou_xy(
            self.Nx,
            self.Ny,
            self.G,
            self.creator.x_list,
            self.creator.y_list,
            self.creator.eps_lists,
        ) * (1.0 + 0.0j)
        self.EPS2 = sub.fou_yx(
            self.Nx,
            self.Ny,
            self.G,
            self.creator.x_list,
            self.creator.y_list,
            self.creator.eps_lists,
        ) * (1.0 + 0.0j)
        return (self.FOUP, self.INV, self.EPS1, self.EPS2)


if __name__ == "__main__":
    fig, ax = plt.subplots(1, 3, figsize=(12, 4))
    cr = Creator()
    cr.rect(12.0, 2.0, 0.5, 0.2)
    lay = Layer(15, 15, cr)
    t = np.linspace(-0.5, 0.5, 101)
    x, y, z = t, t, 0.0
    x, y, z = t, 0.0, t
    x, y, z = 0.0, t, t
    x, y, z = t, t, t
    eps = lay.calculate_epsilon(x, y, z)
    ax[0].contourf(
        eps["x"][:, :, 50],
        eps["y"][:, :, 50],
        eps["eps"][:, :, 50],
    )
    ax[1].contourf(
        eps["x"][:, 50, :],
        eps["z"][:, 50, :],
        eps["eps"][:, 50, :],
    )
    ax[2].contourf(
        eps["y"][50, :, :],
        eps["z"][50, :, :],
        eps["eps"][50, :, :],
    )

    plt.show()



"""Python implementation the Aperiodic-Fourier Modal Method. 
A fully vectorial method for solving Maxwell equations that combines a Fourier-based mode solver and a scattering matrix recursion algorithm to model full 3D structures.
This approach is well suited to calculate modes, transmission, reflection, scattering and absorption of multi-layered structures.
Moreover, support for Bloch modes of periodic structures allows for the simulation of photonic crystals or waveguide Bragg gratings.
"""
import os


from A_FMM.layer import *
from A_FMM.scattering import S_matrix
from A_FMM.creator import Creator
from A_FMM.stack import Stack
import A_FMM.inputs as inputs


from ._version import __version__



import numpy as np


def conv2(x,y,x0,z0,theta):
    #theta=np.pi*theta/180.0
    #z=-x*np.sin(theta)+z0*np.cos(theta)
    z=(x0-x)*np.sin(theta)+z0*np.cos(theta)
    #r=np.abs(x+z0*np.tan(theta))*np.cos(theta)
    r=np.abs(x-x0+z0*np.tan(theta))*np.cos(theta)
    r=np.sqrt(r**2.0+np.cos(theta)**2.0*y**2.0)
    return (z,r)

def FULL_CONV(X,Y,x0,y0,z0,thetax,thetay):
    if thetax==0.0:
        if thetay==0.0:
            phi=0.0
            theta=0.0
        else:
            #phi=np.arctan(np.tan(np.pi/180.0*thetay)/np.tan(np.pi/180.0*thetax))
            #theta=np.arctan(np.tan(np.pi/180.0*thetay)/np.sin(np.pi/180.0*phi))
            phi=np.pi/2.0
            theta=np.pi/180.0*thetay
    else:
        phi=np.arctan(np.tan(np.pi/180.0*thetay)/np.tan(np.pi/180.0*thetax))
        theta=np.arctan(np.tan(np.pi/180.0*thetax)/np.cos(np.pi/180.0*phi))

    #phi=0.5*np.pi-phi

    Xp=np.cos(phi)*(X-x0)-np.sin(phi)*(Y-y0)
    Yp=np.sin(phi)*(X-x0)+np.cos(phi)*(Y-y0)

    x0n=z0*np.tan(theta)

    #Xp=np.cos(phi)*X-np.sin(phi)*Y-x0
    #Yp=np.sin(phi)*X+np.cos(phi)*Y-y0

    (z,r)=conv2(Xp,Yp,x0n,z0,theta)
    return (z,r)

def gauss_beam(r,z,w0,lam):
    zr=w0**2.0/lam*np.pi
    A=1.0+(0.0+1.0j)*z/zr
    return 1.0/A*np.exp(-r**2.0/(w0**2.0*A))*np.exp((0.0+2.0j)*np.pi/lam*z)

def gaussian(X,Y,x0,y0,z0,thetax,thetay,w0,lam):
    (Z,R)=FULL_CONV(X,Y,x0,y0,z0,thetax,thetay)
    G=gauss_beam(R,Z,w0,lam)
    return G




    



from typing import Self

import numpy as np

import A_FMM
import A_FMM.layer
import A_FMM.sub_sm as sub
from A_FMM.layer import Layer
from A_FMM.scattering import S_matrix
import matplotlib.pyplot as plt
import copy
from A_FMM.layer import Layer_empty_st
from matplotlib.backends.backend_pdf import PdfPages

# =============================================================================
# try:
#     from multiprocessing import Pool
# except ModuleNotFoundError:
#     print('WARNING: multiprocessing not available')
# =============================================================================


class Stack:
    """Class representing the multylayer object

    This class is used for the definition of the multilayer object to be simulated using fourier expansion (x and y axis) and scattering matrix algorithm (z axis).
    It is built from a list of layers and thicknesses. The value of the thickness of the first and last layer is irrelevant for the simulation, and it is used only to set the plotting window.
    """

    def __init__(self, layers: list[Layer] = None, d: list[float] = None) -> None:
        """Creator

        Args:
            layers (list, optional): List of Layers: layers of the multylayer. Defaults to None (empty list).
            d (list, optional): List of float: thicknesses of the multylayer. Defaults to None (empty list).

        Raises:
            ValueError: Raised if d and mat have different lengths

        Returns:
            None.

        """
        layers = [] if layers is None else layers
        d = [] if layers is None else d

        if len(layers) != len(d):
            raise ValueError("Different number of layers and thicknesses")
        self.N = len(layers)
        self.layers = layers
        self.d = d
        self.NPW = self.layers[0].D
        self.G = self.layers[0].G
        self.Nyx = self.layers[0].Nyx

        self.count_interface()

    @property
    def total_length(self):
        return sum(self.d)

    def add_layer(self, lay: Layer, d: float) -> None:
        """Add a layer at the end of the multilayer


        Args:
            lay (Layer): Layer to be added.
            d (float): thickness of the layer.

        Returns:
            None.

        """
        self.layers.append(lay)
        self.d.append(d)
        self.N += 1
        self.count_interface()

    def transform(
        self, ex: float = 0, ey: float = 0, complex_transform: bool = False
    ) -> tuple[np.ndarray]:
        """Function for adding the real coordinate transform to all layers of the stack

        Note: for no mapping, set the width to 0

        Args:
            ex (float): relative width of the unmapped region in x direction. Default is 0 (no mapping)
            ey (float): relative width of the unmapped region in y direction. Default is 0 (no mapping)
            complex_transform (bool): False for real transform (default), True for complex one.
        """
        Fx, Fy = self.layers[0].transform(
            ex=ex, ey=ey, complex_transform=complex_transform
        )
        for layer in self.layers[1:]:
            layer.add_transform_matrix(ex=ex, FX=Fx, ey=ey, FY=Fy)
        return Fx, Fy

    def count_interface(self) -> None:
        """Helper function to identify the different layers and the needed interfaces

        Returns:
            None.

        """
        self.tot_thick = sum(self.d)
        self.lay_list = []
        for lay in self.layers:
            if not lay in self.lay_list:
                self.lay_list.append(lay)
        self.int_list = []
        self.interfaces = []
        for i in range(self.N - 1):
            T_inter = (self.layers[i], self.layers[i + 1])
            if not T_inter in self.int_list:
                self.int_list.append(T_inter)
            self.interfaces.append(T_inter)

    def solve(self, k0: float, kx: float = 0.0, ky: float = 0.0) -> None:
        """Calculates the scattering matrix of the multilayer (cpu friendly version)

        This version of solve solve the system in the "smart" way, solving fisrt the eigenvalue problem in each unique layer and the interface matrices of all the interface involved. The computaitonal time scales with the number of different layers, not with the total one.
        It prioritize minimize the calculation done while using more memory.

        Args:
            k0 (float): Vacuum wavevector for the simulation (freqency).
            kx (float, optional): Wavevector in the x direction for the pseudo-fourier expansion. Defaults to 0.0.
            ky (float, optional): Wavevector in the x direction for the pseudo-fourier expansion. Defaults to 0.0.

        Returns:
            None.

        """
        for lay in self.lay_list:
            lay.mode(k0, kx=kx, ky=ky)
            # lay.get_P_norm()
        self.layers[0].get_P_norm()
        self.layers[-1].get_P_norm()
        self.int_matrices = []
        for i in self.int_list:
            self.int_matrices.append(i[0].interface(i[1]))
        self.S = copy.deepcopy(self.int_matrices[0])
        for i in range(1, self.N - 1):
            self.S.add_uniform(self.layers[i], self.d[i])
            self.S.add(self.int_matrices[self.int_list.index(self.interfaces[i])])

    def solve_serial(self, k0: float, kx: float = 0.0, ky: float = 0.0) -> None:
        """Calculates the scattering matrix of the multilayer (memory friendly version)

        This version solves sequentially the layers and the interface as they are in the stack. It is more momery efficient since onlt the data of 2 layer are kept in memory at any given time. Computational time scales with the total number of layer, regardless if they are equal or not.
        It prioritize memory efficiency while possibly requiring more calculations.

        Args:
            k0 (float): Vacuum wavevector for the simulation (freqency).
            kx (float, optional): Wavevector in the x direction for the pseudo-fourier expansion. Defaults to 0.0.
            ky (float, optional): Wavevector in the x direction for the pseudo-fourier expansion. Defaults to 0.0.

        Returns:
            None.

        """
        lay1 = self.layers[0]
        lay1.mode(k0, kx=kx, ky=ky)
        lay1.get_P_norm()
        self.S = S_matrix(2 * self.NPW)
        for i in range(1, self.N):
            lay2 = self.layers[i]
            lay2.mode(k0, kx=kx, ky=ky)
            self.S.add(lay1.interface(lay2))
            self.S.add_uniform(lay2, self.d[i])
            if lay1 is not lay2 and i != 1 and i != self.N:
                lay1.clear()
            lay1 = lay2
        lay2.mode(k0, kx=kx, ky=ky)
        lay2.get_P_norm()

    def solve_lay(self, k0: float, kx: float = 0.0, ky: float = 0.0) -> None:
        """Solve the eigenvalue problem of all the layer in the stack


        Args:
            k0 (float): Vacuum wavevector for the simulation (freqency).
            kx (float, optional): Wavevector in the x direction for the pseudo-fourier expansion. Defaults to 0.0.
            ky (float, optional): Wavevector in the x direction for the pseudo-fourier expansion. Defaults to 0.0.

        Returns:
            None.

        """
        for lay in self.lay_list:
            lay.mode(k0, kx=kx, ky=ky)
            # lay.get_P_norm()
        self.layers[0].get_P_norm()
        self.layers[-1].get_P_norm()

    def solve_S(self) -> None:
        """Builds the scattering matrix of the stacks. It assumes that all the layers are alredy solved.

        Returns:
            None.

        """
        self.int_matrices = []
        for i in self.int_list:
            self.int_matrices.append(i[0].interface(i[1]))
        self.S = copy.deepcopy(self.int_matrices[0])
        for i in range(1, self.N - 1):
            self.S.add_uniform(self.layers[i], self.d[i])
            self.S.add(self.int_matrices[self.int_list.index(self.interfaces[i])])

    def get_prop(
        self, u: np.ndarray, list_lay: list[int], d: np.ndarray = None
    ) -> dict[int, float]:
        """Calculates the total poyinting vector in the requiested layers


        Args:
            u (ndarray): array containing the modal coefficient incoming in the first layer.
            list_lay (list of int): indexes of the layer of which to calculate the Poynting vector.
            d (ndarray, optional): array containing the modal coefficient incoming in the last layer. Defaults to None.

        Returns:
            dic (dict): Dictionary of the Poyting vectors in the the layers {layer_index : Poyting vector}

        """
        dic = {}
        u1, d2 = np.zeros((2 * self.NPW), complex), np.zeros((2 * self.NPW), complex)
        u1 = u
        if d != None:
            d2 = d
        (u2, d1) = self.S.output(u1, d2)
        lay = self.layers[0]
        d = self.d[0]
        if 0 in list_lay:
            P = lay.get_Poynting(u1, d1)
            dic[0] = P
        # intermediate layers
        S1 = copy.deepcopy(self.int_matrices[self.int_list.index(self.interfaces[0])])
        for i in range(1, self.N - 1):
            S2 = S_matrix(S1.N)
            for l in range(i, self.N - 1):
                S2.add_uniform(self.layers[l], self.d[l])
                S2.add(self.int_matrices[self.int_list.index(self.interfaces[l])])
            if i in list_lay:
                (ul, dl) = S1.int_f_tot(S2, u1, d2)
                P = self.layers[i].get_Poynting(ul, dl)
                dic[i] = P
            S1.add_uniform(self.layers[i], self.d[i])
            S1.add(self.int_matrices[self.int_list.index(self.interfaces[i])])
        lay = self.layers[-1]
        d = self.d[-1]
        if self.N - 1 in list_lay:
            P = lay.get_Poynting(u2, d2)
            dic[self.N - 1] = P
        return dic

    def get_energybalance(self, u: np.ndarray, d: np.ndarray = None) -> tuple[float]:
        """Get total energy balance of the stack given the inputs

        Return total power reflected, transmitted and absorbed, normalized to the incidenc power.

        Args:
            u (1darray): Modal coefficient of the left input.
            d (1darray, optional): Modal coefficent of the right input. Defaults to None.

        Returns:
            tuple: tuple contining tree floats with meaning:
                - Total power out from left side (reflection if only u).
                - Total power out from right side (transmission if only u).
                - Total power absorbed in the stack.

        """
        u1, d2, e = (
            np.zeros((2 * self.NPW), complex),
            np.zeros((2 * self.NPW), complex),
            np.zeros((2 * self.NPW), complex),
        )
        u1 = u
        PN = self.layers[0].get_Poynting(u1, e)
        if d is not None:
            d2 = d
            PN -= self.layers[-1].get_Poynting(e, d2)
        (u2, d1) = self.S.output(u1, d2)
        P1 = self.layers[0].get_Poynting(u1, d1)
        P2 = self.layers[-1].get_Poynting(u2, d2)
        return P1 / PN, P2 / PN, (P1 - P2) / PN

    def get_inout(
        self, u: np.ndarray, d: np.ndarray = None
    ) -> dict[str, tuple[np.ndarray, np.ndarray, float]]:
        """Return data about the output of the structure given the input


        Args:
            u (1darray): Vector of the modal coefficents of the right inputs.
            d (1darray, optional): Vector of the modal coefficents of the right inputs. Defaults to None.

        Returns:
            dict: Dictionary containing data of the output:
                - 'left' : (u,d,P): forward modal coefficient, backward modal coefficient and Poyinting vector at the left side.
                - 'right' : (u,d,P): forward modal coefficient, backward modal coefficient and Poyinting vector at the right side.
        """

        u1, d2 = np.zeros((2 * self.NPW), complex), np.zeros((2 * self.NPW), complex)
        u1 = u
        if d != None:
            d2 = d
        (u2, d1) = self.S.output(u1, d2)
        dic = {}
        P = self.layers[0].get_Poynting(u1, d1)
        dic["left"] = (u1, d1, P)
        P = self.layers[-1].get_Poynting(u2, d2)
        dic["right"] = (u2, d2, P)
        return dic

    def get_R(self, i: int, j: int, ordered: bool = True) -> float:
        """Get relfection coefficient between modes

        Args:
            i (int): Index of the source mode.
            j (int): Index of the target mode.
            ordered (bool, optional): If True, modes are ordered for decrasing effective index, otherwise the order is whatever is returned by the diagonalization routine. Defaults to True.

        Returns:
            float: Reflection between the modes

        """
        if ordered:
            j1 = np.argsort(self.layers[0].W)[-i - 1]
            j2 = np.argsort(self.layers[0].W)[-j - 1]
        else:
            j1 = i
            j2 = j
        return (
            np.abs(self.S.S21[j1, j2]) ** 2
            * self.layers[0].P_norm[j2]
            / self.layers[0].P_norm[j1]
        )

    def get_T(self, i: int, j: int, ordered: bool = True) -> float:
        """Get transmission coefficient between modes.

        Args:
            i (int): Index of the source mode.
            j (int): Index of the target mode.
            ordered (bool, optional): If True, modes are ordered for decrasing effective index, otherwise the order is whatever is returned by the diagonalization routine. Defaults to True.

        Returns:
            float: Transmission between the modes.

        """
        if ordered:
            j1 = np.argsort(self.layers[0].W)[-i - 1]
            j2 = np.argsort(self.layers[-1].W)[-j - 1]
        else:
            j1 = i
            j2 = j
        return (
            np.abs(self.S.S11[j2, j1]) ** 2
            * self.layers[-1].P_norm[j2]
            / self.layers[0].P_norm[j1]
        )

    def get_PR(self, i: int, j: int, ordered: bool = True) -> float:
        """Get phase of the relfection coefficient between modes

        Args:
            i (int): Index of the source mode.
            j (int): Index of the target mode.
            ordered (bool, optional): If True, modes are ordered for decrasing effective index, otherwise the order is whatever is returned by the diagonalization routine. Defaults to True.

        Returns:
            float: Phase of reflection between the modes

        """
        if ordered:
            j1 = np.argsort(self.layers[0].W)[-i - 1]
            j2 = np.argsort(self.layers[0].W)[-j - 1]
        else:
            j1 = i
            j2 = j
        return np.angle(self.S.S21[j2, j1])

    def get_PT(self, i: int, j: int, ordered: bool = True) -> float:
        """Get phase of the transmission coefficient between modes

        Args:
            i (int): Index of the source mode.
            j (int): Index of the target mode.
            ordered (bool, optional): If True, modes are ordered for decrasing effective index, otherwise the order is whatever is returned by the diagonalization routine. Defaults to True.

        Returns:
            float: Phase of transmission between the modes

        """
        if ordered:
            j1 = np.argsort(self.layers[0].W)[-i - 1]
            j2 = np.argsort(self.layers[-1].W)[-j - 1]
        else:
            j1 = i
            j2 = j
        return np.angle(self.S.S11[j2, j1])

    def get_el(self, sel: str, i: int, j: int) -> complex:
        """Returns element of the scattering matrix

        Note: Modes are ordered for decrasing effective index

        Args:
            sel (str): First index of the matrix.
            i (int): Second index of the matrix.
            j (int): Select the relevand submatrix. Choises are '11', '12', '21', '22'.

        Raises:
            ValueError: If sel in not in the allowed.

        Returns:
            complex: Element of the scattering matrix.

        """
        io = np.argsort(self.layers[0].W)[-i]
        jo = np.argsort(self.layers[-1].W)[-j]
        if sel == "11":
            return self.S.S11[io, jo]
        elif sel == "12":
            return self.S.S12[io, jo]
        elif sel == "21":
            return self.S.S21[io, jo]
        elif sel == "22":
            return self.S.S22[io, jo]
        else:
            raise ValueError(f"Sel {sel} not allowed. Only '11', '12', '21', '22'")

    def double(self) -> None:
        """Compose the scattering matrix of the stack with itself, doubling the structure

        When doing this, the lenght of the first al last layer are ignored (set to 0).
        To function properly hoever they need to be equal (but do not need to have physical meaning)

        Raises:
            RuntimeError: Raised if the stack is not solved yet.

        Returns:
            None.

        """
        try:
            self.S.add(self.S)
        except AttributeError:
            raise RuntimeError("structure not solved yet")

    def join(self, st2: Self) -> None:
        """Join the scattering matrix of the structure with the one of a second structure

        When doing this, the lenght of the first al last layeror each stack are ignored (set to 0).
        To function last layer of self and first of st2 need to be equal (but do not need to have physical meaning).
        The condiction used to previoselt solve the stack needs to be the same. This is not checked by the code, so be careful.


        Args:
            st2 (Stack): Stack to which to join self.

        Raises:
            RuntimeError: Raised is one the structure is not solved yet.

        Returns:
            None.

        """
        try:
            self.S
        except AttributeError:
            raise RuntimeError("structure 1 not solved yet")
        try:
            st2.S
        except AttributeError:
            raise RuntimeError("structure 2 not solved yet")
        self.S.add(st2.S)
        l1 = self.layers[:-1]
        l2 = st2.layers[1:]
        self.layers = l1 + l2

    def flip(self) -> None:
        """Flip a solved stack

        Flip the stack, swapping the left and right side

        Raises:
            RuntimeError: Raised if the structure is not solved yet.

        Returns:
            None.

        """
        try:
            S = copy.deepcopy(self.S)
            self.S.S11 = S.S22
            self.S.S22 = S.S11
            self.S.S12 = S.S21
            self.S.S21 = S.S12
        except AttributeError:
            raise RuntimeError("structure not solved yet")
        self.layers = self.layers[::-1]
        self.d = self.d[::-1]

    def bloch_modes(self) -> np.ndarray:
        """Calculates Bloch modes of the stack.

        This function assumens the stack to represent the unit cell of a periodic structure, and calculates the corresponding Bloch modes.
        The thickness of the first and last layer are ignored (assumed 0). To work correctly first and last layer needs to be the same.

        Returns:
            TYPE: DESCRIPTION.

        """
        [self.BW, self.BV] = self.S.S_modes()
        self.Bk = -(0.0 + 1j) * np.log(self.BW) / (2.0 * np.pi * self.tot_thick)
        # reorder modes
        ind = np.argsort((0.0 + 1.0j) * self.Bk)
        self.BW = self.BW[ind]
        self.Bk = self.Bk[ind]
        self.BV[:, :] = self.BV[:, ind]
        return self.Bk

    def loop_intermediate(self, u1: np.ndarray, d2: np.ndarray) -> tuple:
        """Generator for the intermedia modal coefficients.

        Progressively yields the forward and backward modal coefficient given the external excitation.

        Args:
            u1 (np.ndarray): forward modal coefficient of the first layer (near the interface)
            d2 ((np.ndarray):  backward modal coefficient of the last layer (near the interface)

        Yields:
            np.ndarray: forward modal amplitudes of layer
            np.ndarray: backward modal amplitudes for layer
            Layer: layer object
            float: thickness of the layer
        """
        u2, d1 = self.S.output(u1, d2)
        lay = self.layers[0]
        d = self.d[0]
        yield u1 * np.exp(-(0 + 2j) * np.pi * lay.k0 * lay.gamma * d), d1 * np.exp(
            (0 + 2j) * np.pi * lay.k0 * lay.gamma * d
        ), self.layers[0], self.d[0]
        # yield u1 , d1 , self.layers[0], self.d[0]
        S1 = copy.deepcopy(self.int_matrices[self.int_list.index(self.interfaces[0])])
        for i in range(1, self.N - 1):
            S2 = S_matrix(S1.N)
            for l in range(i, self.N - 1):
                S2.add_uniform(self.layers[l], self.d[l])
                S2.add(self.int_matrices[self.int_list.index(self.interfaces[l])])
            ul, dl = S1.int_f(S2, u1)
            yield ul, dl, self.layers[i], self.d[i]
            S1.add_uniform(self.layers[i], self.d[i])
            S1.add(self.int_matrices[self.int_list.index(self.interfaces[i])])
        yield u2, d2, self.layers[self.N - 1], self.d[self.N - 1] + 1e-6

    def calculate_epsilon(
        self,
        x: np.ndarray = 0.0,
        y: np.ndarray = 0.0,
        z: np.ndarray = 0.0,
    ) -> dict[str, np.ndarray]:
        """Returns epsilon in the stack

        Epsilon is calculated on a meshgrdi of x,y,z

        Args:
            x (np.ndarray): x coordinate (1D array)
            y (np.ndarray): y coordinate (1D array)
            z (np.ndarray): z coordinate (1D array)

        Returs:
            dict: Dictionary containing the coordinates and the epsilon
        """

        x, y, z = np.asarray(x), np.asarray(y), np.asarray(z)
        eps = {key: [] for key in ["x", "y", "z", "eps"]}
        cumulative_t = 0
        for lay, t in zip(self.layers + [self.layers[-1]], self.d + [np.inf]):
            ind = np.logical_and(cumulative_t <= z, z < cumulative_t + t)
            if ind.size == 0:
                continue
            zp = z[ind] - cumulative_t
            out = lay.calculate_epsilon(x, y, zp)
            out["z"] = out["z"] + cumulative_t
            for key in eps:
                eps[key].append(out[key])
            cumulative_t += t
        for key in eps:
            eps[key] = np.concatenate(eps[key], axis=-1)
        return eps

    def calculate_fields(
        self,
        u1: np.ndarray,
        d2: np.ndarray = None,
        x: np.ndarray = 0,
        y: np.ndarray = 0,
        z: np.ndarray = 0,
        components: list[str] = None,
    ) -> dict[str, np.ndarray]:
        """Returns fields in the stack

        The fields are calculated on a meshgrdi of x,y,z

        Args:
            u1 (np.ndarray): forward modal coefficient in the first layer
            d2 (np.ndarray): backward modal coefficient in the last layer
            x (np.ndarray): x coordinate (1D array)
            y (np.ndarray): y coordinate (1D array)
            z (np.ndarray): z coordinate (1D array)
            components (list): List of modal componets to be calculated. Possible are ['Ex', 'Ey', 'Hx', 'Hz'].
                Default to None (all of them).

        Returs:
            dict: Dictionary containing the coordinates and the field components
        """
        d2 = np.zeros(2 * self.NPW, dtype=complex)
        x, y, z = np.asarray(x), np.asarray(y), np.asarray(z)
        components = Layer._filter_componets(components)
        shape = Layer._check_array_shapes(u1, d2)
        keys = ["x", "y", "z"] + components
        field = {key: [] for key in keys}
        cumulative_t = 0
        for u, d, lay, t in self.loop_intermediate(u1, d2):
            # print(u, d)
            ind = np.logical_and(cumulative_t <= z, z < cumulative_t + t)
            if ind.size == 0:
                continue
            zp = z[ind] - cumulative_t
            out = lay.calculate_field(u, d, x, y, zp, components=components)
            out["z"] = out["z"] + cumulative_t
            for key in keys:
                field[key].append(out[key])
            # plt.plot(np.abs(out['Ex']))
            # plt.show()
            # plt.plot(np.abs(field['Ex']))
            cumulative_t += t
        for key in keys:
            field[key] = np.concatenate(field[key], axis=-1)
        return field

    def inspect(self, st: str = "", details: str = "no") -> None:
        """Print some info about the Stack"""
        att = sub.get_user_attributes(self)
        print(st)
        print(22 * "_")
        print("| INT argument")
        for i in att:
            if type(i[1]) is int:
                print("|%10s%10s" % (i[0], str(i[1])))
        print("| Float argument")
        for i in att:
            if type(i[1]) is float:
                print("|%10s%10s" % (i[0], str(i[1])))
        for i in att:
            if type(i[1]) is np.float64:
                print("|%10s%10s" % (i[0], str(i[1])))
        print("| BOOL argument")
        for i in att:
            if type(i[1]) is bool:
                print("|%10s%10s" % (i[0], str(i[1])))
        print("| Array argument")
        for i in att:
            if type(i[1]) is np.ndarray:
                print("|%10s%10s" % (i[0], str(np.shape(i[1]))))
        print("| List argument")
        for i in att:
            if type(i[1]) is list:
                print("|%12s%8s" % (i[0], str(len(i[1]))))
        print("")
        try:
            print("lay list:")
            for s in self.lay_list:
                print(s)

            print("layers:")
            for s in self.layers:
                print(s)

            print("int_list:")
            for s in self.int_list:
                print(s)

            print("interfaces:")
            for s in self.interfaces:
                print(s)
        except AttributeError:
            print("No list yet, call conut_interface before inspect")


if __name__ == "__main__":
    from monitor import Timer
    import pickle

    timer = Timer()

    N = 50
    cr = A_FMM.Creator()
    cr.slab(12.0, 2.0, 2.0, 0.3)
    lay1 = A_FMM.layer.Layer(N, 0, creator=cr)
    cr.slab(12.0, 2.0, 2.0, 0.1)
    lay2 = A_FMM.layer.Layer(N, 0, creator=cr)

    stack = Stack(
        10 * [lay1, lay2] + [lay1],
        [0.0] + 10 * [0.5, 0.5],
    )

    x, y, z = np.linspace(-0.5, 0.5, 101), 0.0, np.linspace(0.0, 10.0, 1000)
    eps = stack.calculate_epsilon(x, y, z)

    print(eps.keys())
    plt.contourf(
        np.squeeze(eps["z"]), np.squeeze(eps["x"]), np.squeeze(eps["eps"]), levels=41
    )
    plt.show()

    # lay1 = A_FMM.layer.Layer_uniform(0,0,2.0)
    # lay2 = A_FMM.layer.Layer_uniform(0,0,12.0)
    # stack = Stack(
    #     10 * [lay1, lay2] + [lay1],
    #     [0.0] + 10*[0.5, 0.5],
    # )
    # stack.solve(0.1)
    # x, y, z = 0.0, 0.0, np.linspace(0.0, 10.0, 1000)
    # with timer:
    #     field = stack.calculate_fields([1.0, 0.0], [0.0, 0.0], x,y,z)
    # print(timer.elapsed_time)
    # with open('test_stack_1Dfield.pkl', 'wb') as pkl:
    #     pickle.dump(field, pkl, protocol=pickle.HIGHEST_PROTOCOL)

    # layer = lay1.calculate_field([1.0, 0.0], [0.0, 0.0], x,y,z)

    # #plt.contourf(np.squeeze(field['z']), np.squeeze(field['x']), np.abs(np.squeeze(field['Ex'])), levels=41)
    # plt.show()
    # with timer:
    #     Ex, Ey = stack.plot_E(1, func=np.abs, dz = 0.01)
    # print(timer.elapsed_time)
    # Ex = np.asarray(Ex)
    # plt.show()

    # plt.plot(field['z'][0,0,:], np.abs(field['Ex'][0,0,:]))
    # plt.plot(z, np.abs(Ex[:, 50]))
    # print(np.shape(field['z']))
    # plt.show()



import numpy as np
from scipy import integrate
from scipy import linalg
import matplotlib.pyplot as plt
import numpy.linalg
import inspect
from matplotlib.backends.backend_pdf import PdfPages


def savefig(pdf, fig):
    if isinstance(pdf, PdfPages):
        pdf.savefig(fig)
    elif isinstance(pdf, str):
        with PdfPages(pdf) as a:
            a.savefig(fig)
    if pdf is not None:
        plt.close()


def get_user_attributes(cls):
    boring = dir(type("dummy", (object,), {}))
    return [item for item in inspect.getmembers(cls) if item[0] not in boring]


def createG(Ng1, Ng2):
    dic = {}
    n = 0
    for i in range(-Ng1, Ng1 + 1):
        for j in range(-Ng2, Ng2 + 1):
            #    for j in range(-Ng2,Ng2+1):
            #        for i in range(-Ng1,Ng1+1):
            dic[n] = (i, j)
            n += 1
    return dic


def inter_(n, x_list, eps_list):
    if n == 0:
        return eps_list[0] + sum(
            [
                -(eps_list[(j + 1) % len(eps_list)] - eps_list[j]) * x_list[j]
                for j in range(len(x_list))
            ]
        )
    else:
        return sum(
            [
                (eps_list[(j + 1) % len(eps_list)] - eps_list[j])
                * np.exp(-(0 + 2j) * np.pi * n * x_list[j])
                for j in range(len(x_list))
            ]
        ) / ((0 + 2j) * np.pi * n)


def inter_v(n, x_list, eps_list):
    if n == 0:
        return 1.0 / eps_list[0] + sum(
            [
                -(1.0 / eps_list[(j + 1) % len(eps_list)] - 1.0 / eps_list[j])
                * x_list[j]
                for j in range(len(x_list))
            ]
        )
    else:
        return sum(
            [
                (1.0 / eps_list[(j + 1) % len(eps_list)] - 1.0 / eps_list[j])
                * np.exp(-(0 + 2j) * np.pi * n * x_list[j])
                for j in range(len(x_list))
            ]
        ) / ((0 + 2j) * np.pi * n)


def fou(nx, ny, x_list, y_list, eps_lists):
    N = len(x_list)
    f = []
    for i in range(N):
        f.append(inter_(ny, y_list, eps_lists[i]))
    return (1 + 0j) * inter_(nx, x_list, f)


def create_epsilon(G, x_list, y_list, eps_lists):
    D = len(G)
    F = np.zeros((D, D), complex)
    for i in range(D):
        for j in range(D):
            F[i, j] = fou(
                G[i][0] - G[j][0], G[i][1] - G[j][1], x_list, y_list, eps_lists
            )
    return F


def fou_v(nx, ny, x_list, y_list, eps_lists):
    N = len(x_list)
    f = []
    for i in range(N):
        f.append(inter_v(ny, y_list, eps_lists[i]))
    return (1 + 0j) * inter_(nx, x_list, f)


def fou_yx(Nx, Ny, G, x_list, y_list, eps_lists):
    f = []
    D = len(G)
    nx = len(x_list)
    for i in range(nx):
        f.append(
            linalg.inv(
                linalg.toeplitz(
                    [inter_v(j + Ny, y_list, eps_lists[i]) for j in range(-Ny, Ny + 1)]
                )
            )
        )
    F = np.zeros((D, D), complex)
    mx = 4 * Nx + 1
    f = [inter_((i + 2 * Nx) % mx - 2 * Nx, x_list, f) for i in range(mx)]
    for i in range(D):
        for j in range(D):
            F[i, j] = f[G[i][0] - G[j][0]][G[i][1] + Ny, G[j][1] + Ny]
    return F


def fou_xy(Nx, Ny, G, x_list, y_list, eps_lists):
    f = []
    D = len(G)
    nx = len(x_list)
    ny = len(y_list)
    for i in range(ny):
        eps_t = [eps_lists[j][i] for j in range(nx)]
        f.append(
            linalg.inv(
                linalg.toeplitz(
                    [inter_v(j + Nx, x_list, eps_t) for j in range(-Nx, Nx + 1)]
                )
            )
        )
    F = np.zeros((D, D), complex)
    my = 4 * Ny + 1
    f = [inter_((i + 2 * Ny) % my - 2 * Ny, y_list, f) for i in range(my)]
    for i in range(D):
        for j in range(D):
            F[i, j] = f[G[i][1] - G[j][1]][G[i][0] + Nx, G[j][0] + Nx]
    return F


def plot(n_max, x_list, eps_list, N=100):
    x = np.linspace(-0.5, 0.5, N)
    y = np.zeros(N, complex)
    for n in range(-n_max, n_max + 1):
        y += inter_(n, x_list, eps_list) * np.exp((0 + 2j) * np.pi * n * x)
    #        y+= slab_f(n,0.4)*np.exp((0+2j)*np.pi*n*x)

    plt.plot(x, np.real(y))
    plt.show()


def createK(dic, k0, kx=0.0, ky=0.0, Nyx=1.0):
    D = len(dic)
    K1 = np.zeros((D, D), dtype=complex)
    K2 = np.zeros((D, D), dtype=complex)
    for i in range(D):
        #        K1[i,i]=2.0*np.pi*dic[i][0]/k0*(1+0j)
        #        K2[i,i]=2.0*np.pi*dic[i][1]/k0*(1+0j)

        K1[i, i] = (1 + 0j) * (dic[i][0] + kx) / k0
        K2[i, i] = (1 + 0j) * (dic[i][1] + ky) / k0 / Nyx
    return (K1, K2)


def create_2order(dic, K1, K2, INV, EPS1, EPS2):
    D = len(dic)
    ID = np.identity(D, dtype="complex")
    I11 = -np.dot(K2, np.dot(INV, K1))
    I12 = np.dot(K2, np.dot(INV, K2)) - ID
    I21 = ID - np.dot(K1, np.dot(INV, K1))
    I22 = np.dot(K1, np.dot(INV, K2))
    B12 = np.vstack([np.hstack([I11, I12]), np.hstack([I21, I22])])
    I11 = np.dot(K1, K2)
    I22 = -np.dot(K1, K2)
    I21 = np.dot(K1, K1) - EPS2
    I12 = EPS1 - np.dot(K2, K2)
    B21 = np.vstack([np.hstack([I11, I12]), np.hstack([I21, I22])])
    #    B=np.dot(B12,B21)
    #    print 3*'%15.8e' % (numpy.linalg.cond(B12),numpy.linalg.cond(B21),numpy.linalg.cond(B))
    return np.dot(B12, B21)


# convention of Li article
def create_2order_new(D, Kx, Ky, INV, EPS1, EPS2):
    ID = np.identity(D, dtype="complex")
    F = np.vstack(
        [
            np.hstack([np.dot(np.dot(Kx, INV), Ky), ID - np.dot(np.dot(Kx, INV), Kx)]),
            np.hstack([np.dot(np.dot(Ky, INV), Ky) - ID, -np.dot(np.dot(Ky, INV), Kx)]),
        ]
    )
    G = np.vstack(
        [
            np.hstack([-np.dot(Kx, Ky), np.dot(Kx, Kx) - EPS2]),
            np.hstack([EPS1 - np.dot(Ky, Ky), np.dot(Kx, Ky)]),
        ]
    )
    # old version for control
    # G=np.vstack([np.hstack([-np.dot(Kx,Ky),np.dot(Kx,Kx)-EPS2]),np.hstack([EPS1-np.dot(Ky,Ky),np.dot(Ky,Kx)])])
    # F=np.vstack([np.hstack([np.dot(np.dot(Kx,INV),Ky),ID-np.dot(np.dot(Kx,INV),Kx)]),np.hstack([np.dot(np.dot(Ky,INV),Ky)-ID,-np.dot(np.dot(Ky,INV),Kx)])])
    return G, np.dot(F, G)


# convention of Lalanne article
# def create_2order_new(D,Kx,Ky,INV,EPS1,EPS2):
#    ID=np.identity(D,dtype='complex')
#    F=np.vstack([np.hstack([np.dot(np.dot(Ky,INV),Kx),ID-np.dot(np.dot(Ky,INV),Ky)]),np.hstack([np.dot(np.dot(Kx,INV),Kx)-ID,-np.dot(np.dot(Kx,INV),Ky)])])
#    G=np.vstack([np.hstack([np.dot(Kx,Ky),EPS1-np.dot(Ky,Ky)]),np.hstack([np.dot(Kx,Kx)-EPS2,-np.dot(Kx,Ky)])])
#    return np.dot(F,G)


def fou_t(n, e):
    q = 1.0 - e
    return 1.0 * (n == 0) - 0.5 * q * (-1) ** n * (
        np.sinc(n * q) + 0.5 * np.sinc(n * q - 1) + 0.5 * np.sinc(n * q + 1)
    )


def t_inv(x, e):
    q = 1.0 - e
    e = 0.5 * e
    xr = np.copy(x)
    ind = np.abs(x) > e
    xr[ind] = np.sign(x[ind]) * (
        e + q / np.pi * np.arctan(np.pi / q * (np.abs(x[ind]) - e))
    )
    return xr


def t_dir(x, e):
    q = 1.0 - e
    e = 0.5 * e
    xr = np.copy(x)
    ind = np.abs(x) > e
    xr[ind] = np.sign(x[ind]) * (
        e + q / np.pi * np.tan(np.pi / q * (np.abs(x[ind]) - e))
    )
    return xr


def fou_complex_t(n, e):
    g = 1.0 / (1 - 1j)
    q = 1.0 - e
    return 1.0 * (n == 0) - 0.5 * q * (-1) ** n * (
        (1.0 + 0.25 * g) * np.sinc(n * q)
        + 0.5 * np.sinc(n * q - 1)
        + 0.5 * np.sinc(n * q + 1)
        - g / 8.0 * (np.sinc(n * q - 2) + np.sinc(n * q + 2))
    )


def num_fou(func, args, G, NX, NY, Nyx):
    # [X,Y]=np.meshgrid(np.linspace(-0.5,0.5,NX),np.linspace(-0.5,0.5,NY))
    [Y, X] = np.meshgrid(np.linspace(-0.5, 0.5, NY), np.linspace(-0.5, 0.5, NX))
    F = func(X, Y / Nyx, *args)
    F = np.fft.fftshift(F)
    FOU = np.fft.fft2(F) / NX / NY
    EPS = np.zeros((len(G), len(G)), dtype=complex)
    for i in range(len(G)):
        for j in range(len(G)):
            EPS[i, j] = FOU[G[i][0] - G[j][0], G[i][1] - G[j][1]]
    return EPS


def num_fou_xy(func, args, nx, ny, G, NX, NY, Nyx):
    # [X,Y]=np.meshgrid(np.linspace(-0.5,0.5,NX),np.linspace(-0.5,0.5,NY))
    [Y, X] = np.meshgrid(np.linspace(-0.5, 0.5, NY), np.linspace(-0.5, 0.5, NX))
    F = 1.0 / func(X, Y / Nyx, *args)
    F = np.fft.fftshift(F)
    np.shape(F)
    FOU = np.fft.fft(F, axis=0) / NX
    # plt.figure()
    # plt.imshow(np.abs(F),origin='lower')
    # plt.colorbar()
    # plt.savefig('F.png')
    # plt.figure()
    # plt.imshow(np.abs(FOU[:,:20]),aspect='auto',origin='lower')
    # plt.colorbar()
    # plt.savefig('FOU.png')
    TEMP1 = np.zeros((NY, 2 * nx + 1, 2 * nx + 1), dtype=complex)
    for i in range(-nx, nx + 1):
        for j in range(-nx, nx + 1):
            TEMP1[:, i, j] = FOU[i - j, :]
    # print TEMP1[900,:,:]
    TEMP2 = np.linalg.inv(TEMP1)
    TEMP3 = np.fft.fft(TEMP2, axis=0) / NY
    EPS = np.zeros((len(G), len(G)), dtype=complex)
    for i in range(len(G)):
        for j in range(len(G)):
            EPS[i, j] = TEMP3[G[i][1] - G[j][1], G[i][0], G[j][0]]
    return EPS
    # return None


def num_fou_yx(func, args, nx, ny, G, NX, NY, Nyx):
    # [X,Y]=np.meshgrid(np.linspace(-0.5,0.5,NX),np.linspace(-0.5,0.5,NY))
    [Y, X] = np.meshgrid(np.linspace(-0.5, 0.5, NY), np.linspace(-0.5, 0.5, NX))
    F = 1.0 / func(X, Y / Nyx, *args)
    F = np.fft.fftshift(F)
    FOU = np.fft.fft(F, axis=1) / NY
    # plt.figure()
    # plt.imshow(np.abs(F),origin='lower')
    # plt.colorbar()
    # plt.savefig('F.png')
    # plt.figure()
    # plt.imshow(np.abs(FOU[:,:20]),aspect='auto',origin='lower')
    # plt.colorbar()
    # plt.savefig('FOU.png')
    TEMP1 = np.zeros((NX, 2 * ny + 1, 2 * ny + 1), dtype=complex)
    for i in range(-ny, ny + 1):
        for j in range(-ny, ny + 1):
            TEMP1[:, i, j] = FOU[:, i - j]
    # print TEMP1[900,:,:]
    TEMP2 = np.linalg.inv(TEMP1)
    TEMP3 = np.fft.fft(TEMP2, axis=0) / NX
    EPS = np.zeros((len(G), len(G)), dtype=complex)
    for i in range(len(G)):
        for j in range(len(G)):
            EPS[i, j] = TEMP3[G[i][0] - G[j][0], G[i][1], G[j][1]]
    return EPS
    # return None



import numpy as np
from scipy import linalg
import matplotlib.pyplot as plt
import A_FMM.sub_sm as sub
from matplotlib.backends.backend_pdf import PdfPages


class Creator:
    """Class for the definition of the eps profile in the layer"""

    def __init__(self, x_list=None, y_list=None, eps_lists=None):
        """Creator

        Args:
            x_list (list):      list of floats containig the coordinates of the x boundaries
            y_list (list):      list of floats containig the coordinates of the y boundaries
            eps_lists (list):   list of list of floats containig the eps value of the squares defined by x_list and y_list
        """
        self.x_list = x_list
        self.y_list = y_list
        self.eps_lists = eps_lists

    def slow_general(self, eps_core, eps_lc, eps_uc, w, h, t, Z):
        self.x_list = np.linspace(-0.5 * w, 0.5 * w, len(Z) + 1)
        self.y_list = [-0.5, -0.5 * h, -0.5 * h + t, 0.5 * h]
        self.eps_lists = [[eps_uc, eps_lc, eps_core, eps_uc]]
        eps = [eps_uc, eps_core]
        for i in Z:
            self.eps_lists.append([eps_uc, eps_lc, eps_core, eps[i]])

    def slow_2D(self, eps_core, eps_c, w, Z):
        self.x_list = np.linspace(-0.5 * w, 0.5 * w, len(Z) + 1)
        self.y_list = [-0.5]
        self.eps_lists = [[eps_c]]
        eps = [eps_c, eps_core]
        for i in Z:
            self.eps_lists.append([eps[i]])

    def ridge(self, eps_core, eps_lc, eps_uc, w, h, t=0.0, y_offset=0.0, x_offset=0.0):
        """Rib waveguide with single layer

        Args:
            eps_core (float):   epsilon of the core
            eps_lc (floar):     epsilon of the lower cladding
            eps_up (float):     epsilon of the upper cladding
            w (float):          width of the rib (in unit of ax)
            h (float):          height of the un-etched part (in unit of ay)
            t (float):          height of the etched part (in unit of ay). Default is 0 (strip waveguide)
            x_offset (float):   offset of the center of the waveguide with respec to the center of the cell (in unit of ax). Default is 0
            y_offset (float):   offset of the etched part with resoect to the unetched one (in unit of ay). Default is 0 (etched and unetched part are aligned at the bottom)
        """
        self.x_list = [-0.5 * w + x_offset, 0.5 * w + x_offset]
        self.y_list = [
            -0.5,
            -0.5 * h,
            -0.5 * h + y_offset,
            -0.5 * h + t + y_offset,
            0.5 * h,
        ]
        self.eps_lists = [
            [eps_uc, eps_lc, eps_lc, eps_core, eps_uc],
            [eps_uc, eps_lc, eps_core, eps_core, eps_core],
        ]

    def ridge_pn(self, eps0, epsp, epsn, eps_lc, eps_uc, w, h, t, xp, xn):
        if xp < -0.5 * w:
            x_left = [xp, -0.5 * w]
            eps_left = [[eps_uc, eps_lc, epsp, eps_uc], [eps_uc, eps_lc, eps0, eps_uc]]
        else:
            x_left = [-0.5 * w, xp]
            eps_left = [[eps_uc, eps_lc, epsp, eps_uc], [eps_uc, eps_lc, epsp, epsp]]
        if xn > 0.5 * w:
            x_right = [0.5 * w, xn, 0.5]
            eps_right = [
                [eps_uc, eps_lc, eps0, eps0],
                [eps_uc, eps_lc, eps0, eps_uc],
                [eps_uc, eps_lc, epsn, eps_uc],
            ]
        else:
            x_right = [xn, 0.5 * w, xn]
            eps_right = [
                [eps_uc, eps_lc, eps0, eps0],
                [eps_uc, eps_lc, epsn, epsn],
                [eps_uc, eps_lc, epsn, eps_uc],
            ]

        self.x_list = x_left + x_right
        self.y_list = [-0.5, -0.5 * h, -0.5 * h + t, 0.5 * h]
        self.eps_lists = eps_left + eps_right

    def ridge_double(
        self, eps_core, eps_lc, eps_uc, w1, w2, h, t1, t2, y_offset=0.0, x_offset=0.0
    ):
        """Rib waveguide with double etch

        Args:
            eps_core (float):   epsilon of the core
            eps_lc (floar):     epsilon of the lower cladding
            eps_up (float):     epsilon of the upper cladding
            w1 (float):         width of the unetched part (in unit of ax)
            w2 (float):         width of the intermediate etched part (in unit of ax)
            h (float):          height of the un-etched part (in unit of ay)
            t1 (float):         height of the intermidiate etched part (in unit of ay).
            t2 (float):         height of the maximum etched part (in unit of ay).
            x_offset (float):   offset of the center of the waveguide with respec to the center of the cell (in unit of ax). Default is 0
            y_offset (float):   offset of the etched part with resoect to the unetched one (in unit of ay). Default is 0 (etched and unetched part are aligned at the bottom)

        """

        self.x_list = [
            -0.5 * w2 + x_offset,
            -0.5 * w1 + x_offset,
            0.5 * w1 + x_offset,
            0.5 * w2 + x_offset,
        ]
        self.y_list = [
            -0.5,
            -0.5 * h,
            -0.5 * h + y_offset,
            -0.5 * h + t2 + y_offset,
            -0.5 * h + t1 + y_offset,
            0.5 * h,
        ]
        self.eps_lists = [
            [eps_uc, eps_lc, eps_lc, eps_core, eps_uc, eps_uc],
            [eps_uc, eps_lc, eps_lc, eps_core, eps_core, eps_uc],
            [eps_uc, eps_lc, eps_core, eps_core, eps_core, eps_core],
            [eps_uc, eps_lc, eps_lc, eps_core, eps_core, eps_uc],
        ]

    def rect(self, eps_core, eps_clad, w, h, off_x=0.0, off_y=0.0):
        """Rectangular waveguide

        Args:
            eps_core (float):   epsilon of the core
            eps_clad (floar):   epsilon of the cladding
            w (float):          width of the waveguide (in unit of ax)
            h (float):          height of the waveguide (in unit of ay)
            off_y (float):      offset of the center of the waveguide with respect to the cell (in unit of ay). Default is 0.
            off_x (float):      offset of the center of the waveguide with respect to the cell (in unit of ax). Default is 0.
        """
        self.x_list = [-0.5 * w + off_x, 0.5 * w + off_x]
        self.y_list = [-0.5 * h + off_y, 0.5 * h + off_y]
        self.eps_lists = [[eps_clad, eps_clad], [eps_clad, eps_core]]

    def slab(self, eps_core, eps_lc, eps_uc, w, offset=0.0):
        """1D slab in x direction

        Args:
            eps_core (float): epsilon of the core.
            eps_lc (float): epsilon of the lower cladding.
            eps_uc (float): epsilon of the upper cladding.
            w (float): thickness of the slab (in unit of ax).
            offset (float, optional): Offset if the slab with respect to the center of the cell. Defaults to 0.0.

        Returns:
            None.

        """
        self.x_list = [-0.5, -0.5 * w + offset, 0.5 * w + offset]
        self.y_list = [0.5]
        self.eps_lists = [[eps_uc], [eps_lc], [eps_core]]

    def slab_y(self, eps_core, eps_lc, eps_uc, w):
        """1D slab in y direction

        Args:
            eps_core (float): epsilon of the core.
            eps_lc (float): epsilon of the lower cladding.
            eps_uc (float): epsilon of the upper cladding.
            w (float): thickness of the slab (in unit of ay).
            offset (float, optional): Offset if the slab with respect to the center of the cell. Defaults to 0.0.

        Returns:
            None.

        """
        self.x_list = [0.5]
        self.y_list = [-0.5, -0.5 * w, 0.5 * w]
        self.eps_lists = [[eps_uc, eps_lc, eps_core]]

    def x_stack(self, x_l, eps_l):
        self.y_list = [0.5]
        self.x_list = [-0.5] + x_l
        self.eps_lists = [[eps_l[-1]]]
        for eps in eps_l:
            self.eps_lists.append([eps])

    def hole(self, h, w, r, e_core, e_lc, e_up, e_fill):
        """Rib waveguide with a hole in the middle


        Args:
            h (TYPE): height of the waveguide (in unit of ay).
            w (TYPE): width of the waveguide (in unit of ax).
            r (TYPE): radius of the internal hole (in unit of ax).
            e_core (TYPE): epsilon of the core.
            e_lc (TYPE): epsilon of the lower cladding.
            e_up (TYPE): epsilon of the upper cladding.
            e_fill (TYPE): epsilon inside the hole.

        Returns:
            None.

        """
        self.x_list = [-0.5 * w, -r, r, 0.5 * w]
        self.y_list = [-0.5 * h, 0.5 * h, 0.5]
        self.eps_lists = [
            [e_lc, e_up, e_up],
            [e_lc, e_core, e_up],
            [e_lc, e_fill, e_up],
            [e_lc, e_core, e_up],
        ]

    def circle(self, e_in, e_out, r, n):
        self.x_list = np.linspace(-r, r, n)
        self.y_list = np.linspace(-r, r, n)
        [X, Y] = np.meshgrid(self.x_list, self.y_list)
        # ind= np.sqrt((X-0.5*r/float(n))**2+(Y-0.5*r/float(n))**2)<r
        ind = np.sqrt(X**2 + Y**2) < r
        # eps=np.array([e_out,e_in])
        self.eps_lists = e_out + ind * (e_in - e_out)

    def etched_stack(self, eps_uc, eps_lc, w, etch, eps_stack, d_stack):
        h = sum(d_stack)
        self.x_list = [-0.5 * w, 0.5 * w]
        self.y_list = [-0.5]
        eps1 = [eps_uc, eps_lc]
        eps2 = [eps_uc, eps_lc]
        dd = np.cumsum(d_stack)
        if etch > h:
            self.y_list.append(0.5 * h - etch)
            eps1.append(eps1[-1])
            eps2.append(eps_uc)
            dec = 1
        else:
            dec = 0
        for d, eps in zip(reversed(dd), reversed(eps_stack)):
            if (d < etch) and (dec == 0):
                self.y_list.append(0.5 * h - etch)
                eps1.append(eps1[-1])
                eps2.append(eps_uc)
                dec = 1
            self.y_list.append(0.5 * h - d)
            eps1.append(eps)
            if dec == 0:
                eps2.append(eps)
            else:
                eps2.append(eps_uc)
        self.y_list.append(0.5 * h)
        self.eps_lists = [eps2, eps1]

    def varied_epi(self, eps_back, data_list, y_off=0.0):
        t_tot = sum([dd[2] for dd in data_list])
        w_list = np.sort(list(set([0.5 * dd[1] for dd in data_list])))
        self.x_list = [-0.5] + list(-w_list[::-1]) + list(w_list)
        self.y_list = [-0.5 * t_tot] + list(
            -0.5 * t_tot + np.cumsum([dd[2] for dd in data_list])
        )
        self.y_list = [_ + y_off for _ in self.y_list]
        self.eps_lists = [len(self.y_list) * [eps_back]]
        for pos in self.x_list:
            eps_list = [eps_back]
            for eps, w, t in data_list:
                if pos < -0.5 * w:
                    eps_list.append(eps_back)
                elif pos >= 0.5 * w:
                    eps_list.append(eps_back)
                else:
                    eps_list.append(eps)
            self.eps_lists.append(eps_list)

    def varied_plane(self, eps_back, t, data_list):
        self.y_list = [-0.5 * t, 0.5 * t]
        w_tot = sum([dd[1] for dd in data_list])
        self.x_list = [-0.5 * w_tot] + list(
            -0.5 * w_tot + np.cumsum([dd[1] for dd in data_list])
        )
        self.eps_lists = [len(self.y_list) * [eps_back]]
        for i, (eps, w) in enumerate(data_list):
            eps_list = [eps_back, eps]
            self.eps_lists.append(eps_list)

        # print(self.y_list)
        # print(self.x_list)
        # print(self.eps_lists)

    def plot_eps(self, N=101):
        EPS = np.zeros((N, N)) + self.eps_lists[0][0]
        x = np.linspace(-0.5, 0.5, N)
        y = np.linspace(-0.5, 0.5, N)
        x, y = np.meshgrid(x, y, indexing="ij")
        x_old = -0.5
        for xv, el in zip(self.x_list, self.eps_lists):
            EPS[np.logical_and(x >= xv, x <= x_old)] = el[0]
            for yv, e in zip(self.y_list, el):
                EPS[np.logical_and(x >= xv, y >= yv)] = e
                plt.imshow(np.logical_and(x <= xv, x >= x_old, y >= yv))
                plt.show()
                x_old = xv
        return EPS



import numpy as np
import scipy.linalg as linalg

class S_matrix:
    """Implementation of the scattring matrix object
    
    This object is a container for NxN matrices, conventionally defined as S11, S12, S21 and S22
    Also, it implementens all the methods involving operations on scattring matrix
    """

    def __init__(self,N):
        """Creator
        
        Args:
            N (int): Dimension of each of the NxN submatrices of the scattring matrix. The total matrix is 2Nx2N

        Returns:
            None.

        """
        self.N=N
        self.S11=np.identity(N,complex)
        self.S22=np.identity(N,complex)
        self.S12=np.zeros((N,N),complex)
        self.S21=np.zeros((N,N),complex)


    #OLD RECURSION VERSION
    #def add(self,s):
    #    T1=np.dot(linalg.inv(np.identity(self.N,complex)-np.dot(self.S12,s.S21)),self.S11)
    #    T2=np.dot(linalg.inv(np.identity(self.N,complex)-np.dot(s.S21,self.S12)),s.S22)
    #    self.S11=np.dot(s.S11,T1)
    #    self.S12=s.S12+np.dot(np.dot(s.S11,self.S12),T2)
    #    self.S21=self.S21+np.dot(np.dot(self.S22,s.S21),T1)
    #    self.S22=np.dot(self.S22,T2)

    #NEW RECURSION VERSION
    def add(self,s):
        """Recursion method for joining two scattering matrices
        
        The connection is between the "right" side of self and the "left" side of s
        
        Args:
            s (S_matrix): scattering matrix to be joined to self. The 

        Returns:
            None.

        """
        I=np.identity(self.N,complex)
        T1=np.dot(s.S11,linalg.inv(I-np.dot(self.S12,s.S21)))
        T2=np.dot(self.S22,linalg.inv(I-np.dot(s.S21,self.S12)))
        self.S21=self.S21+np.dot(np.dot(T2,s.S21),self.S11)
        self.S11=np.dot(T1,self.S11)
        self.S12=s.S12   +np.dot(np.dot(T1,self.S12),s.S22)             
        self.S22=np.dot(T2,s.S22)

    def add_left(self,s):
        """Recursion method for joining two scattering matrices
        
        The connection is between the "left" side of self and the "right" side of s
        
        Args:
            s (S_matrix): scattering matrix to be joined to self. The 

        Returns:
            None.

        """
        T1=np.dot(linalg.inv(np.identity(self.N,complex)-np.dot(s.S12,self.S21)),s.S11)
        T2=np.dot(linalg.inv(np.identity(self.N,complex)-np.dot(self.S12,s.S21)),self.S22)
        s.S11=np.dot(self.S11,T1)
        s.S12=self.S12+np.dot(np.dot(self.S11,s.S12),T2)
        s.S21=s.S21+np.dot(np.dot(s.S22,self.S21),T1)
        s.S22=np.dot(s.S22,T2)

    def add_uniform(self,lay,d):
        """Recursion method for addig to self the progation matrix of a given layer
        
        The connection is between the "right" side of self and the "left" side of the propagation matrix  

        Args:
            lay (Layer): Layer of which to calculate the propagation matrix
            d (float): Thickness of the layer

        Returns:
            None.

        """
        E=np.diag(np.exp((0+2j)*np.pi*lay.k0*lay.gamma*d))
        self.S11=np.dot(E,self.S11)
        self.S12=np.dot(E,np.dot(self.S12,E))
        self.S22=np.dot(self.S22,E)

    def add_uniform_left(self,lay,d):
        """Recursion method for addig to self the progation matrix of a given layer
        
        The connection is between the "left" side of self and the "right" side of the propagation matrix  

        Args:
            lay (Layer): Layer of which to calculate the propagation matrix
            d (float): Thickness of the layer

        Returns:
            None.

        """
        E=np.diag(np.exp((0+2j)*np.pi*lay.k0*lay.gamma*d))
        self.S11=np.dot(self.S11,E)
        self.S21=np.dot(E,np.dot(self.S21,E))
        self.S22=np.dot(E,self.S22)


    def S_print(self,i=None,j=None):
        """Function for printing the scattering matrix. 
        
        It can print both the full matrix or the 2x2 matrix between relevant modes

        Args:
            i (int, optional): index of the "left" mode. Default is None (full matrix)
            j (int, optional): index of the "right" mode. Default is None (full matrix)

        Returns:
            None.

        """
        if i==None:
            S=np.vstack([np.hstack([self.S11,self.S12]),np.hstack([self.S21,self.S22])])
        else:
            j=i if j==None else j
            S=np.vstack([np.hstack([self.S11[i,j],self.S12[i,j]]),np.hstack([self.S21[i,j],self.S22[i,j]])])
        print(S)

    def det(self):
        """Calculate the determinat of the scattering matrix

        Returns:
            float: Determinant of the scattering matrix

        """
        return linalg.det(np.vstack([np.hstack([self.S11,self.S12]),np.hstack([self.S21,self.S22])]))

    def S_modes(self):
        """Solves the eigenvalue problem of the Bloch modes of the scattring matrix
        
        Returns:
            W (1darray): arrays of the eigenvalues (complex amplitude of the mode after one period)
            V (2darray):  arrays of the eigenvectors (Bloch modes on the base of the guided mode in the first and last layer)

        """
        ID=np.identity(self.N)
        Z=np.zeros((self.N,self.N))
        S1=np.vstack([np.hstack([self.S11,Z]),np.hstack([self.S21,-ID])])
        S2=np.vstack([np.hstack([ID,-self.S12]),np.hstack([Z,-self.S22])])
        [W,V]=linalg.eig(S1,b=S2)
        return W,V

    def det_modes(self,kz,d):
        ID=np.identity(self.N)
        Z=np.zeros((self.N,self.N))
        S1=np.vstack([np.hstack([self.S11,Z]),np.hstack([self.S21,-ID])])
        S2=np.vstack([np.hstack([ID,-self.S12]),np.hstack([Z,-self.S22])])
        return linalg.det(S1-np.exp((0.0+1.0j)*kz*d)*S2)        

    def der(self,Sm,Sp,h=0.01):
        """Calculates the first and second derivative of the scattering matrix with respec to the parameter par.

        Args:
            Sm (S_matrix): S matrix calculated at par=par0-h
            Sp (S_matrix): S matrix calculated at par=par0+h
            h (float, optional): Interval used to calculate the derivatives . Defaults to 0.01.

        Returns:
            tuple: tuple containing:
                
                - S1 (2darray): First derivative of the scattering matrix with respect to par.
                - S2 (2darray): Second derivative of the scattering matrix with respect to par.

        """
        S=np.vstack([np.hstack([self.S11,self.S12]),np.hstack([self.S21,self.S22])])
        S_m=np.vstack([np.hstack([Sm.S11,Sm.S12]),np.hstack([Sm.S21,Sm.S22])])
        S_p=np.vstack([np.hstack([Sp.S11,Sp.S12]),np.hstack([Sp.S21,Sp.S22])])
        S1=(S_p-S_m)/(2.0*h)
        S2=(S_p+S_m-2.0*S)/(h*h)
        return (S1,S2)

    def matrix(self):
        """Returns the full scattering matrix
        
        Returns:
            2darray: Scattering matrix as numpy array

        """
        return np.vstack([np.hstack([self.S11,self.S12]),np.hstack([self.S21,self.S22])])

    def output(self,u1,d2):
        """Returs the output vectors given the input vectors

        Args:
            u1 (1darray): Array of modal coefficient of "left" inputs.
            d2 (1darray): Array of modal coefficient of "right" inputs.

        Returns:
            tuple: tuple containing:
                
                - u2 (1darray): Array of modal coefficient of "right" outputs.
                - d1 (1darray): Array of modal coefficient of "left" outputs.

        """
        u2=np.add(np.dot(self.S11,u1),np.dot(self.S12,d2))
        d1=np.add(np.dot(self.S21,u1),np.dot(self.S22,d2))
        return (u2,d1)

    def left(self,u1,d1):
        """Return the "right" inout and output vectors given the "left" ones
        

        Args:
            u1 (1darray): Array of modal coefficient of "left" inputs.
            d1 (1darray): Array of modal coefficient of "left" outputs.

        Returns:
            tuple: tuple containing:
                
                - u2 (1darray): Array of modal coefficient of "right" outputs.
                - d2 (1darray): Array of modal coefficient of "right" inputs.

        """
        d2=linalg.solve(self.S22,d1-np.dot(self.S21,u1))
        u2=np.add(np.dot(self.S11,u1),np.dot(self.S21,d2))
        return (u2,d2)

    def int_f(self,S2,u):
        """Retirn the modal coefficient between two scattering matrces (self and S2)

        Args:
            S2 (S_matrix): Scattering matrix to between self and the end of the structure
            u (1darray): Array of modal coefficient of "left" inputs to self.

        Returns:
            tuple: tuple containing:
                
                - uo (TYPE): Array of coefficients of left-propagating modes in the middle 
                - do (TYPE): Array of coefficients of right-propagating modes in the middle

        """
        ID=np.identity(self.N)
        ut=np.dot(self.S11,u)
        uo=linalg.solve(ID-np.dot(self.S12,S2.S21),ut)
        do=linalg.solve(ID-np.dot(S2.S21,self.S12),np.dot(S2.S21,ut))
        return (uo,do)

    def int_f_tot(self,S2,u,d):
        """Retirn the modal coefficient between two scattering matrces (self and S2)

        Args:
            S2 (S_matrix): Scattering matrix to between self and the end of the structure
            u (1darray): Array of modal coefficient of "left" inputs to self.
            d (1darray): Array of modal coefficient of "right" inputs to S2

        Returns:
            tuple: tuple containing:
                
                - uo (TYPE): Array of coefficients of left-propagating modes in the middle 
                - do (TYPE): Array of coefficients of right-propagating modes in the middle

        """
        ID=np.identity(self.N)
        ut=np.dot(self.S11,u)
        dt=np.dot(S2.S22,d)
        uo=linalg.solve(ID-np.dot(self.S12,S2.S21),np.add(ut,np.dot(self.S12,dt)))
        do=linalg.solve(ID-np.dot(S2.S21,self.S12),np.add(np.dot(S2.S21,ut),dt))
        return (uo,do)

    int_complete = int_f_tot
            




